@@ -1,21 +1,54 @@
-# Copyright (c) 2019 NVIDIA Corporation
-__all__ = ['Backend',
-           'ModelMode',
-           'Optimization',
-           'DeviceType',
-           'Actions',
-           'NeuralModuleFactory']
+# ! /usr/bin/python
+# -*- coding: utf-8 -*-
+
+# Copyright (c) 2019, NVIDIA CORPORATION. All rights reserved.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
+__all__ = [
+    'Backend',
+    'ModelMode',
+    'Optimization',
+    'DeviceType',
+    'Actions',
+    'NeuralModuleFactory',
+    'DeploymentFormat',
+]
 
-from abc import ABC, abstractmethod
 import random
+from abc import ABC, abstractmethod
+from enum import Enum
 from typing import List, Optional
 
-from enum import Enum
 import numpy as np
 
+import nemo
+from ..utils import ExpManager
 from .callbacks import ActionCallback, EvaluatorCallback
 from .neural_types import *
-from ..utils import ExpManager
+from nemo.utils.decorators import deprecated
+
+logging = nemo.logging
+
+
+class DeploymentFormat(Enum):
+    """Which format to use when exporting a Neural Module for deployment"""
+
+    AUTO = 0
+    PYTORCH = 1
+    TORCHSCRIPT = 2
+    ONNX = 3
+    TRTONNX = 4
 
 
 class Backend(Enum):
@@ -53,9 +86,9 @@ class DeviceType(Enum):
 class Actions(ABC):
     """Basic actions allowed on graphs of Neural Modules"""
 
-    def __init__(self, local_rank,
-                 optimization_level=Optimization.mxprO0):
+    def __init__(self, local_rank, global_rank, optimization_level=Optimization.mxprO0):
         self._local_rank = local_rank
+        self._global_rank = global_rank
         self._optim_level = optimization_level
         self.step = None
         self.epoch_num = None
@@ -69,14 +102,23 @@ class Actions(ABC):
         """
         return self._local_rank
 
+    @property
+    def global_rank(self):
+        """Global rank during distributed execution. None if single GPU/CPU
+
+        Returns:
+            (int) rank or worker or None if not in distributed model
+        """
+        return self._global_rank
+
     @abstractmethod
     def train(
-            self,
-            tensors_to_optimize: List[NmTensor],
-            callbacks: Optional[List[ActionCallback]],
-            lr_policy=None,
-            batches_per_step=None,
-            stop_on_nan_loss=False
+        self,
+        tensors_to_optimize: List[NmTensor],
+        callbacks: Optional[List[ActionCallback]],
+        lr_policy=None,
+        batches_per_step=None,
+        stop_on_nan_loss=False,
     ):
         """This action executes training and (optionally) evaluation.
 
@@ -93,9 +135,9 @@ class Actions(ABC):
                 "algorithmic" batch size per GPU/worker = batches_per_step*
                 batch_size
             stop_on_nan_loss: (default: False) If set to True, the training
-                will stop if loss=nan. If set to False, the training will
-                continue, but the gradients will be zeroed before next
-                mini-batch.
+                will stop if loss=nan or inf. If set to False, the training
+                will continue. Note that if apex.amp is not used, or if
+                optimization level is O0, training will stop regardless.
 
         Returns:
             None
@@ -138,11 +180,7 @@ class Actions(ABC):
         pass
 
     @abstractmethod
-    def create_optimizer(
-            self,
-            optimizer,
-            things_to_optimize,
-            optimizer_params):
+    def create_optimizer(self, optimizer, things_to_optimize, optimizer_params):
         """
         Creates an optimizer object to be use in the train() method.
 
@@ -160,61 +198,45 @@ class Actions(ABC):
     def _perform_on_iteration_start(self, callbacks):
         # TODO: Most of these checks can be relaxed since we enforce callbacks
         # to be a list of ActionCallback objects
-        if callbacks is not None and isinstance(callbacks, List) and len(
-                callbacks) > 0:
+        if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
             for callback in callbacks:
-                callback._local_rank = self.local_rank
                 callback.on_iteration_start()
 
     def _perform_on_iteration_end(self, callbacks):
-        if callbacks is not None and isinstance(callbacks, List) and len(
-                callbacks) > 0:
+        if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
             for callback in callbacks:
-                callback._local_rank = self.local_rank
                 callback.on_iteration_end()
 
     def _perform_on_action_start(self, callbacks):
-        if callbacks is not None and isinstance(callbacks, List) and len(
-                callbacks) > 0:
+        if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
             for callback in callbacks:
-                callback._local_rank = self.local_rank
                 callback.on_action_start()
 
     def _perform_on_action_end(self, callbacks):
-        if callbacks is not None and isinstance(callbacks, List) and len(
-                callbacks) > 0:
+        if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
             for callback in callbacks:
-                callback._local_rank = self.local_rank
                 callback.on_action_end()
 
     def _perform_on_epoch_start(self, callbacks):
-        if callbacks is not None and isinstance(callbacks, List) and len(
-                callbacks) > 0:
+        if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
             for callback in callbacks:
-                callback._local_rank = self.local_rank
                 callback.on_epoch_start()
 
     def _perform_on_epoch_end(self, callbacks):
-        if callbacks is not None and isinstance(callbacks, List) and len(
-                callbacks) > 0:
+        if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
             for callback in callbacks:
-                callback._local_rank = self.local_rank
                 callback.on_epoch_end()
 
     def _init_callbacks(self, callbacks):
-        if callbacks is not None and isinstance(callbacks, List) and len(
-                callbacks) > 0:
+        if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
             for callback in callbacks:
                 callback.action = self
 
     def _update_callbacks(
-            self,
-            callbacks=None,
-            registered_tensors=None,
+        self, callbacks=None, registered_tensors=None,
     ):
         # if self.local_rank is None or self.local_rank == 0:
-        if callbacks is not None and isinstance(callbacks, List) and len(
-                callbacks) > 0:
+        if callbacks is not None and isinstance(callbacks, List) and len(callbacks) > 0:
             for callback in callbacks:
                 callback._registered_tensors = registered_tensors
 
@@ -256,22 +278,23 @@ class NeuralModuleFactory(object):
     """
 
     def __init__(
-            self,
-            backend=Backend.PyTorch,
-            local_rank=None,
-            optimization_level=Optimization.mxprO0,
-            placement=None,
-            cudnn_benchmark=False,
-            random_seed=None,
-            set_default=True,
-            log_dir=None,
-            checkpoint_dir=None,
-            tensorboard_dir=None,
-            create_tb_writer=False,
-            files_to_copy=None,
-            add_time_to_log_dir=False
+        self,
+        backend=Backend.PyTorch,
+        local_rank=None,
+        optimization_level=Optimization.mxprO0,
+        placement=None,
+        cudnn_benchmark=False,
+        random_seed=None,
+        set_default=True,
+        log_dir=None,
+        checkpoint_dir=None,
+        tensorboard_dir=None,
+        create_tb_writer=False,
+        files_to_copy=None,
+        add_time_to_log_dir=False,
     ):
         self._local_rank = local_rank
+        self._global_rank = None
 
         if isinstance(optimization_level, str):
             optimization_level = _str_to_opt_level(optimization_level)
@@ -289,23 +312,26 @@ class NeuralModuleFactory(object):
 
         self._backend = backend
         self._world_size = 1
+        broadcast_func = None
         if backend == Backend.PyTorch:
             # TODO: Move all framework specific code from this file
             import torch
+
             if self._placement != DeviceType.CPU:
                 if not torch.cuda.is_available():
-                    raise ValueError("You requested to use GPUs but CUDA is "
-                                     "not installed. You can try running using"
-                                     " CPU-only. To do this, instantiate your"
-                                     " factory with placement=DeviceType.CPU"
-                                     "\n"
-                                     "Note that this is slow and is not "
-                                     "well supported.")
+                    raise ValueError(
+                        "You requested to use GPUs but CUDA is "
+                        "not installed. You can try running using"
+                        " CPU-only. To do this, instantiate your"
+                        " factory with placement=DeviceType.CPU"
+                        "\n"
+                        "Note that this is slow and is not "
+                        "well supported."
+                    )
 
             torch.backends.cudnn.benchmark = cudnn_benchmark
             if random_seed is not None and cudnn_benchmark:
-                raise ValueError("cudnn_benchmark can not be set to True"
-                                 "when random_seed is not None.")
+                raise ValueError("cudnn_benchmark can not be set to True when random_seed is not None.")
             if random_seed is not None:
                 torch.backends.cudnn.deterministic = True
                 torch.backends.cudnn.benchmark = False
@@ -313,18 +339,63 @@ class NeuralModuleFactory(object):
                 np.random.seed(random_seed)
                 random.seed(random_seed)
 
-            if self._local_rank is not None:
-                torch.cuda.set_device(self._local_rank)
-                torch.distributed.init_process_group(
-                    backend="nccl", init_method="env://"
-                )
-                self._world_size = torch.distributed.get_world_size()
-        else:
-            raise NotImplementedError(
-                "Only Pytorch backend is currently supported.")
+            # logging.info("Random seeds")
+            # logging.info("torch: %d", torch.initial_seed())
+            # logging.info("numpy: %d", )
+            # logging.info("random: %d", )
 
-        if set_default:
-            NeuralModuleFactory.set_default_factory(self)
+            if self._local_rank is not None:
+                torch.distributed.init_process_group(backend="nccl", init_method="env://")
+
+                cuda_set = True
+                # Try to set cuda device. This should fail if self._local_rank
+                # is greater than the number of available GPUs
+                try:
+                    torch.cuda.set_device(self._local_rank)
+                except RuntimeError:
+                    # Note in this case, all tensors are now sent to GPU 0
+                    # who could crash because of OOM. Thus init_process_group()
+                    # must be done before any cuda tensors are allocated
+                    cuda_set = False
+                cuda_set_t = torch.cuda.IntTensor([cuda_set])
+
+                # Do an all_reduce to ensure all workers obtained a GPU
+                # For the strangest reason, BAND doesn't work so I am resorting
+                # to MIN.
+                torch.distributed.all_reduce(cuda_set_t, op=torch.distributed.ReduceOp.MIN)
+                if cuda_set_t.item() == 0:
+                    raise RuntimeError(
+                        "There was an error initializing distributed training."
+                        " Perhaps you specified more gpus than you have "
+                        "available"
+                    )
+
+                del cuda_set_t
+                torch.cuda.empty_cache()
+                # Remove test tensor from memory
+
+                self._world_size = torch.distributed.get_world_size()
+                self._global_rank = torch.distributed.get_rank()
+
+                def torch_broadcast_wrapper(str_len=None, string=None, src=0):
+                    """Wrapper function to broadcast string values across all
+                    workers
+                    """
+                    # Create byte cuda torch tensor
+                    if string is not None:
+                        string_tensor = torch.tensor(list(string.encode()), dtype=torch.uint8).cuda()
+                    else:
+                        string_tensor = torch.tensor([0] * str_len, dtype=torch.uint8).cuda()
+                    # Run broadcast
+                    torch.distributed.broadcast(string_tensor, src)
+                    # turn byte tensor back to string
+                    return_string = string_tensor.cpu().numpy()
+                    return_string = b''.join(return_string).decode()
+                    return return_string
+
+                broadcast_func = torch_broadcast_wrapper
+        else:
+            raise NotImplementedError("Only Pytorch backend is currently supported.")
 
         # Create ExpManager
         # if log_dir is None, only create logger
@@ -334,14 +405,20 @@ class NeuralModuleFactory(object):
             use_tb=create_tb_writer,
             tb_dir=tensorboard_dir,
             local_rank=local_rank,
+            global_rank=self._global_rank,
             files_to_copy=files_to_copy,
             add_time=add_time_to_log_dir,
-            exist_ok=True)
+            exist_ok=True,
+            broadcast_func=broadcast_func,
+        )
         self._tb_writer = self._exp_manager.tb_writer
 
         # Create trainer
         self._trainer = self._get_trainer(tb_writer=self._tb_writer)
 
+        if set_default:
+            NeuralModuleFactory.set_default_factory(self)
+
     @classmethod
     def get_default_factory(cls):
         return cls._DEFAULT
@@ -362,40 +439,30 @@ class NeuralModuleFactory(object):
             mod = getattr(mod, comp)
         return mod
 
-    def __get_pytorch_module(self, name, params, collection, pretrained):
-        params["factory"] = self
-        if collection == "toys" or collection == "tutorials" or collection \
-                == "other":
-            constructor = NeuralModuleFactory.__name_import(
-                "nemo.backends.pytorch.tutorials." + name
-            )
+    @deprecated(version=0.11)
+    def __get_pytorch_module(self, name, collection, params, pretrained):
+        # TK: "factory" is not passed as parameter anymore.
+        # params["factory"] = self
+
+        if collection == "toys" or collection == "tutorials" or collection == "other":
+            constructor = NeuralModuleFactory.__name_import("nemo.backends.pytorch.tutorials." + name)
         elif collection == "nemo_nlp":
-            constructor = NeuralModuleFactory.__name_import(
-                "nemo_nlp." + name
-            )
+            constructor = NeuralModuleFactory.__name_import("nemo_nlp." + name)
             if name == "BERT" and pretrained is True:
                 params["pretrained"] = True
         elif collection == "nemo_asr":
-            constructor = NeuralModuleFactory.__name_import(
-                "nemo_asr." + name
-            )
+            constructor = NeuralModuleFactory.__name_import("nemo_asr." + name)
         elif collection == "nemo_lpr":
-            constructor = NeuralModuleFactory.__name_import(
-                "nemo_lpr." + name
-            )
+            constructor = NeuralModuleFactory.__name_import("nemo_lpr." + name)
         elif collection == 'common':
-            constructor = NeuralModuleFactory.__name_import(
-                'nemo.backends.pytorch.common.' + name
-            )
+            constructor = NeuralModuleFactory.__name_import('nemo.backends.pytorch.common.' + name)
         elif collection == "torchvision":
             import torchvision.models as tv_models
             import nemo.backends.pytorch.module_wrapper as mw
             import torch.nn as nn
 
             if name == "ImageFolderDataLayer":
-                constructor = NeuralModuleFactory.__name_import(
-                    "nemo.backends.pytorch.torchvision.data." + name
-                )
+                constructor = NeuralModuleFactory.__name_import("nemo.backends.pytorch.torchvision.data." + name)
                 instance = constructor(**params)
                 return instance
             else:
@@ -411,21 +478,14 @@ class NeuralModuleFactory(object):
                             }
                         )
                     }
-                    output_ports = {
-                        "output": NeuralType(
-                            {0: AxisType(BatchTag), 1: AxisType(ChannelTag)}
-                        )
-                    }
+                    output_ports = {"output": NeuralType({0: AxisType(BatchTag), 1: AxisType(ChannelTag)})}
 
                     pt_model = tv_models.resnet18(pretrained=pretrained)
                     num_classes = params.get("num_classes", None)
                     if num_classes is not None:
                         pt_model.fc = nn.Linear(512, params["num_classes"])
                     return mw.TrainableNeuralModuleWrapper(
-                        pt_nn_module=pt_model,
-                        input_ports_dict=input_ports,
-                        output_ports_dict=output_ports,
-                        **params,
+                        pt_nn_module=pt_model, input_ports_dict=input_ports, output_ports_dict=output_ports,
                     )
                 elif _nm_name == "resnet50":
                     input_ports = {
@@ -438,21 +498,14 @@ class NeuralModuleFactory(object):
                             }
                         )
                     }
-                    output_ports = {
-                        "output": NeuralType(
-                            {0: AxisType(BatchTag), 1: AxisType(ChannelTag)}
-                        )
-                    }
+                    output_ports = {"output": NeuralType({0: AxisType(BatchTag), 1: AxisType(ChannelTag)})}
 
                     pt_model = tv_models.resnet50(pretrained=pretrained)
                     num_classes = params.get("num_classes", None)
                     if num_classes is not None:
                         pt_model.fc = nn.Linear(2048, params["num_classes"])
                     return mw.TrainableNeuralModuleWrapper(
-                        pt_nn_module=pt_model,
-                        input_ports_dict=input_ports,
-                        output_ports_dict=output_ports,
-                        **params,
+                        pt_nn_module=pt_model, input_ports_dict=input_ports, output_ports_dict=output_ports,
                     )
         else:
             collection_path = "nemo.collections." + collection + "." + name
@@ -460,12 +513,14 @@ class NeuralModuleFactory(object):
             if name == "BERT" and pretrained is True:
                 params["pretrained"] = True
 
-        if "placement" not in params:
-            params["placement"] = self._placement
+        # TK: "placement" is not passed as parameter anymore.
+        # if "placement" not in params:
+        #    params["placement"] = self._placement
         instance = constructor(**params)
         return instance
 
-    def get_module(self, name, params, collection, pretrained=False):
+    @deprecated(version=0.11)
+    def get_module(self, name, collection, params, pretrained=False):
         """
         Creates NeuralModule instance
 
@@ -481,51 +536,48 @@ class NeuralModuleFactory(object):
         Returns:
           NeuralModule instance
         """
-        if params is not None and "optimization_level" in params:
-            if params["optimization_level"] != self._optim_level:
-                self.logger.warning(
-                    "Module's {0} requested optimization level {1} is"
-                    "different from the one specified by factory - {2}."
-                    "Using: {3} for this module".format(
-                        name,
-                        params["optimization_level"],
-                        self._optim_level,
-                        params["optimization_level"],
-                    )
-                )
-        else:
-            if params is None:
-                params = {}
-            params["optimization_level"] = self._optim_level
+
+        # TK: "optimization_level" is not passed as parameter anymore.
+        # if params is not None and "optimization_level" in params:
+        #    if params["optimization_level"] != self._optim_level:
+        #        logging.warning(
+        #            "Module's {0} requested optimization level {1} is"
+        #            "different from the one specified by factory - {2}."
+        #            "Using: {3} for this module".format(
+        #                name, params["optimization_level"], self._optim_level, params["optimization_level"],
+        #            )
+        #        )
+        # else:
+        #    if params is None:
+        #        params = {}
+        #    params["optimization_level"] = self._optim_level
 
         if self._backend == Backend.PyTorch:
-            return self.__get_pytorch_module(
-                name=name, params=params, collection=collection,
-                pretrained=pretrained
-            )
+            return self.__get_pytorch_module(name=name, collection=collection, params=params, pretrained=pretrained,)
         else:
             return None
 
-    def create_optimizer(self,
-                         optimizer,
-                         things_to_optimize,
-                         optimizer_params):
+    def create_optimizer(self, optimizer, things_to_optimize, optimizer_params):
         return self._trainer.create_optimizer(
-            optimizer=optimizer,
-            things_to_optimize=things_to_optimize,
-            optimizer_params=optimizer_params)
+            optimizer=optimizer, things_to_optimize=things_to_optimize, optimizer_params=optimizer_params,
+        )
 
-    def train(self,
-              tensors_to_optimize,
-              optimizer=None,
-              optimization_params=None,
-              callbacks: Optional[List[ActionCallback]] = None,
-              lr_policy=None,
-              batches_per_step=None,
-              stop_on_nan_loss=False,
-              synced_batchnorm=False,
-              synced_batchnorm_groupsize=0,
-              reset=False):
+    def train(
+        self,
+        tensors_to_optimize,
+        optimizer=None,
+        optimization_params=None,
+        callbacks: Optional[List[ActionCallback]] = None,
+        lr_policy=None,
+        batches_per_step=None,
+        stop_on_nan_loss=False,
+        steps_per_nan_check=100,
+        synced_batchnorm=False,
+        synced_batchnorm_groupsize=0,
+        gradient_predivide=False,
+        amp_max_loss_scale=2.0 ** 24,
+        reset=False,
+    ):
         if reset:
             self.reset_trainer()
         return self._trainer.train(
@@ -536,33 +588,56 @@ class NeuralModuleFactory(object):
             lr_policy=lr_policy,
             batches_per_step=batches_per_step,
             stop_on_nan_loss=stop_on_nan_loss,
+            steps_per_nan_check=steps_per_nan_check,
             synced_batchnorm=synced_batchnorm,
-            synced_batchnorm_groupsize=synced_batchnorm_groupsize)
-
-    def eval(self,
-             callbacks: List[EvaluatorCallback]):
-        if callbacks is None or len(callbacks) == 0:
-            raise ValueError(f"You need to provide at lease one evaluation"
-                             f"callback to eval")
-        for callback in callbacks:
-            if not isinstance(callback, EvaluatorCallback):
-                raise TypeError(f"All callbacks passed to the eval action must"
-                                f"be inherited from EvaluatorCallback")
-        self.train(
-            tensors_to_optimize=None,
-            optimizer='sgd',
-            callbacks=callbacks,
-            optimization_params={'num_epochs': 1}
+            synced_batchnorm_groupsize=synced_batchnorm_groupsize,
+            gradient_predivide=gradient_predivide,
+            amp_max_loss_scale=amp_max_loss_scale,
         )
 
-    def infer(self,
-              tensors: List[NmTensor],
-              checkpoint_dir=None,
-              ckpt_pattern='',
-              verbose=True,
-              cache=False,
-              use_cache=False,
-              offload_to_cpu=True):
+    def eval(self, callbacks: List[EvaluatorCallback]):
+        if callbacks is None or len(callbacks) == 0:
+            raise ValueError(f"You need to provide at lease one evaluation" f"callback to eval")
+        for callback in callbacks:
+            if not isinstance(callback, EvaluatorCallback):
+                raise TypeError(f"All callbacks passed to the eval action must" f"be inherited from EvaluatorCallback")
+        self.train(
+            tensors_to_optimize=None, optimizer='sgd', callbacks=callbacks, optimization_params={'num_epochs': 1},
+        )
+
+    def deployment_export(
+        self, module, output: str, d_format: DeploymentFormat, input_example=None, output_example=None
+    ):
+        """Exports Neural Module instance for deployment.
+
+        Args:
+            module: neural module to export
+            output (str): where export results should be saved
+            d_format (DeploymentFormat): which deployment format to use
+            input_example: sometimes tracing will require input examples
+            output_example: Should match inference on input_example
+        """
+        module._prepare_for_deployment()
+
+        return self._trainer.deployment_export(
+            module=module,
+            output=output,
+            d_format=d_format,
+            input_example=input_example,
+            output_example=output_example,
+        )
+
+    def infer(
+        self,
+        tensors: List[NmTensor],
+        checkpoint_dir=None,
+        ckpt_pattern='',
+        verbose=True,
+        cache=False,
+        use_cache=False,
+        offload_to_cpu=True,
+        modules_to_restore=None,
+    ):
         """Runs inference to obtain values for tensors
 
         Args:
@@ -583,6 +658,9 @@ class NeuralModuleFactory(object):
                 `tensors` list. Defaults to False.
             offload_to_cpu (bool): If True, all evaluated tensors are moved to
                 cpu memory after each inference batch. Defaults to True.
+            modules_to_restore (list): Defaults to None, in which case all
+                NMs inside callchain with weights will be restored. If
+                specified only the modules inside this list will be restored.
 
         Returns:
             List of evaluated tensors. Each element in the list is also a list
@@ -593,10 +671,11 @@ class NeuralModuleFactory(object):
             checkpoint_dir=checkpoint_dir,
             ckpt_pattern=ckpt_pattern,
             verbose=verbose,
-            logger=self.logger,
             cache=cache,
             use_cache=use_cache,
-            offload_to_cpu=offload_to_cpu)
+            offload_to_cpu=offload_to_cpu,
+            modules_to_restore=modules_to_restore,
+        )
 
     def clear_cache(self):
         """Helper function to clean inference cache."""
@@ -604,27 +683,28 @@ class NeuralModuleFactory(object):
 
     def _get_trainer(self, tb_writer=None):
         if self._backend == Backend.PyTorch:
-            constructor = NeuralModuleFactory.__name_import(
-                "nemo.backends.pytorch.PtActions"
+            constructor = NeuralModuleFactory.__name_import("nemo.backends.pytorch.PtActions")
+            instance = constructor(
+                local_rank=self._local_rank,
+                global_rank=self._global_rank,
+                tb_writer=tb_writer,
+                optimization_level=self._optim_level,
             )
-            instance = constructor(local_rank=self._local_rank,
-                                   tb_writer=tb_writer,
-                                   optimization_level=self._optim_level)
             return instance
         else:
             raise ValueError("Only PyTorch backend is currently supported.")
 
+    @deprecated(
+        version="future",
+        explanation="Please use .train(...), .eval(...), .infer(...) and "
+        f".create_optimizer(...) of the NeuralModuleFactory instance directly.",
+    )
     def get_trainer(self, tb_writer=None):
-        self.logger.warning(
-            f"This function is deprecated and will be removed"
-            f"in future versions of NeMo."
-            f"Please use .train(...), .eval(...), .infer(...) and "
-            f".create_optimizer(...) directly methods from "
-            f"NeuralModuleFactory instance.")
         if self._trainer:
-            self.logger.warning(
+            logging.warning(
                 "The trainer instance was created during initialization of "
-                "Neural factory, using the already created instance.")
+                "Neural factory, using the already created instance."
+            )
             return self._trainer
         return self._get_trainer(tb_writer)
 
@@ -632,6 +712,34 @@ class NeuralModuleFactory(object):
         del self._trainer
         self._trainer = self._get_trainer(tb_writer=self._tb_writer)
 
+    def sync_all_processes(self, status=True):
+        """ Helper function for testing that allows proccess 0 to inform all
+        other processes of failures. Does nothing if not using distributed
+        training. Usage example can be seen in examples/asr/jasper_an4.py
+
+        Args:
+            status (bool): Defaults to True. If any proccess passes False, it
+                will trigger a graceful exit on all other processes. It is
+                assumed that the process that passed False will print an error
+                message on its own and exit
+        """
+        if self._world_size == 1:
+            logging.info("sync_all_processes does nothing if there is one process")
+            return
+        if self._backend == Backend.PyTorch:
+            import torch
+
+            status_tensor = torch.cuda.IntTensor([status])
+            torch.distributed.all_reduce(status_tensor, op=torch.distributed.ReduceOp.MIN)
+            if status_tensor.item() == 0:
+                logging.error("At least one process had a failure")
+                if status:
+                    raise ValueError(
+                        f"Process with global rank {self._global_rank} entered"
+                        " sync_all_processes with a passing status, but "
+                        "another process indicated a failure"
+                    )
+
     @property
     def world_size(self):
         return self._world_size
@@ -649,8 +757,9 @@ class NeuralModuleFactory(object):
         return self._optim_level
 
     @property
+    @deprecated(version=0.11, explanation="Please use ``nemo.logging instead``")
     def logger(self):
-        return self._exp_manager.logger
+        return nemo.logging
 
     @property
     def checkpoint_dir(self):
@@ -659,3 +768,7 @@ class NeuralModuleFactory(object):
     @property
     def work_dir(self):
         return self._exp_manager.work_dir
+
+    @property
+    def global_rank(self):
+        return self._global_rank
